﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace JTShtuff
{
    public class JTActionsBehaviour : MonoBehaviour
    {
        public PuppetBehaviour Puppet { get; set; } = null;
        public Dive dive { get; set; } = new Dive();
        public Prone prone { get; set; } = new Prone();
        public static float OGSlowmotionTimescale { get; set; } = Global.main.SlowmotionTimescale;
        public static float SlowmotionTimescale { get; set; } = 0.5f;


        public void FixedUpdate()
        {
            if (!Puppet.enabled) return;

            if (dive.state  != DiveState.ready)  dive.Go();
            if (prone.state != ProneState.ready) prone.Go();
        }

        public void Start()
        {
            dive.Puppet  = this.Puppet;
            prone.Puppet = this.Puppet;
        }


        public static void MaxPayne(bool turnOn = true, float timeScale = 0f)
        {
            if (turnOn)
            {
                if (timeScale == 0) timeScale = SlowmotionTimescale;

                Global.main.SlowmotionTimescale = SlowmotionTimescale;

                if (!Global.main.SlowMotion) Global.main.ToggleSlowmotion();

                return;
            }

            if (Global.main.SlowMotion) Global.main.ToggleSlowmotion();

            Global.main.SlowmotionTimescale = OGSlowmotionTimescale;
        }


       
        public class Prone
        {
            public int frame { get; set; } = 0;
            public ProneState state { get; set; } = ProneState.ready;
            public PuppetBehaviour Puppet { get; set; } = null;
            public float facing = 1;

            public void Init()
            {
                frame = 0;
                
                if (state == ProneState.ready && Puppet.PBO.OverridePoseIndex != Puppet.PuppetPose["JT_Prone"])
                {
                    frame = 0;

                    state = ProneState.start;


                    facing = Puppet.facingLeft ? -1 : 1;

                    Puppet.RB2["UpperBody"].AddForce(Vector2.up * 5 * Puppet.TotalWeight);
                }

            }

            public void Go()
            {
                frame++;

                if (state == ProneState.start) 
                {
                    Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Prone"];
                    
                    Puppet.RB2["UpperBody"].AddForce(Vector2.up * 25 * Puppet.TotalWeight);
                    
                    state = ProneState.drop;
                    frame = 0;

                    Puppet.isProne     = true;
                    Puppet.isCrouching = false;
                    return;
                }

                if(state == ProneState.drop) { 

                    Puppet.RB2["UpperBody"].AddForce(Vector2.right * facing * 10 * Puppet.TotalWeight);
                    Puppet.RB2["LowerLeg"].AddForce(Vector2.left * facing * 5 * Puppet.TotalWeight);
                    Puppet.RB2["LowerLegFront"].AddForce(Vector2.left * facing * 5 * Puppet.TotalWeight);
                    Puppet.RB2["MiddleBody"].AddTorque(-facing * 5 * Puppet.TotalWeight);
                    Puppet.RB2["Foot"].AddForce(Vector2.up * 5 * Puppet.TotalWeight);

                    //if (frame > 2) { 
                    //    Puppet.RB2["LowerBody"].AddForce(Vector2.down * 1 * Puppet.TotalWeight);
                    //}

                    if (frame >= 5)
                    {
                        state = ProneState.final;
                        frame = 0;
                        return;
                    }
                }

                if (state == ProneState.final)
                {
                    state = ProneState.ready;
                }


            }
        }

        public class Dive
        {
            public int frame { get; set; } = 0;
            public float force { get; set; } = 0;
            public DiveState state { get; set; } = DiveState.ready;
            public PuppetBehaviour Puppet { get; set; } = null;
            public float facing { get; set; } = 1;

            public void Init()
            {

                if (state == DiveState.ready)
                {
                    frame = 0;

                    //SetLimbSnapshot();
                    //Strengthen();

                    Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Dive_1"];

                    state = DiveState.start;
                    force = 0;
                    frame = 0;

                    facing = Puppet.facingLeft ? 1 : -1;
                    if (KB.Modifier) MaxPayne(true);
                    //if (KB.Modifier) gameObject.AddComponent<MaxPayne>();
                    
                }
            }

            public void Go()
            {
                frame++;
                Vector2 direction = Puppet.RB2["UpperBody"].position;
                direction.y = 4.0f;
                direction.x = 10 * -facing;

                if (state == DiveState.start)
                {

                    Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Dive_2"];
                    Puppet.RB2["UpperBody"].AddForce(direction * (20 * Puppet.TotalWeight));
                    
                    Puppet.RB2["LowerArm"].AddForce(Vector2.up *  (20 * Puppet.TotalWeight));
                    Puppet.RB2["LowerArmFront"].AddForce(Vector2.up *  (20 * Puppet.TotalWeight));
                    Puppet.RB2["Foot"].AddForce(Vector2.down *  (20 * Puppet.TotalWeight));
                    Puppet.RB2["FootFront"].AddForce(Vector2.down * (20 * Puppet.TotalWeight));

                    state = DiveState.predive;
                    frame = 0;
                    return;
                
                }
                
                if (state == DiveState.predive) {

                    Puppet.RB2["LowerBody"].AddForce(direction * Puppet.TotalWeight);

                    //if (!Puppet.LB["LowerArm"].IsOnFloor && !Puppet.LB["LowerArmFront"].IsOnFloor) {
                    //    Puppet.RB2["LowerArm"].AddForce(Vector2.right * (2 * Puppet.TotalWeight));
                    //    Puppet.RB2["LowerArmFront"].AddForce(Vector2.right * (2 * Puppet.TotalWeight));
                    //    Puppet.RB2["UpperArm"].AddForce(Vector2.right * (2 * Puppet.TotalWeight));
                    //    Puppet.RB2["UpperArmFront"].AddForce(Vector2.right * (2 * Puppet.TotalWeight));
                    //    Puppet.RB2["MiddleBody"].AddTorque(0.7f * -facing * Puppet.TotalWeight);
                    //}

                    if (frame > 14)
                    {
                        Puppet.PBO.OverridePoseIndex = -1;
                        state = DiveState.landed;
                    }
                    //Puppet.RB2["LowerBody"].AddTorque(facing * 10 * Puppet.TotalWeight);
                    //Puppet.RB2["UpperLeg"].AddForce(xdirection * 10 * Puppet.TotalWeight);
                    //Puppet.RB2["UpperLegFront"].AddForce(xdirection * 10 * Puppet.TotalWeight);


                    //Puppet.RB2["UpperBody"].AddForce(xdirection * 25 * Puppet.TotalWeight);
                    //Puppet.RB2["MiddleBody"].AddTorque(-facing * (4 * Puppet.TotalWeight));

                    //Puppet.RB2["LowerLeg"].AddForce(Vector2.up * 15 * Puppet.TotalWeight);
                    //Puppet.RB2["LowerLegFront"].AddForce(Vector2.up  * 5 * Puppet.TotalWeight);
                }

                //if (state == DiveState.predive && frame < 5)
                //{
                //    frame = 0;
                //    state = DiveState.diving;

                //    Puppet.PBO.OverridePoseIndex = Puppet.PuppetPose["JT_Duck"];

                //    Puppet.RB2["UpperBody"].AddRelativeForce(xdirection * (5 * Puppet.TotalWeight));
                //    Puppet.RB2["MiddleBody"].AddTorque(facing * (2 * Puppet.TotalWeight));

                //    Puppet.DisableMoves = Time.time + 2.5f;
                //    return;

                //}

                if (state == DiveState.landed)
                {

                    if (Puppet.LB["Foot"].IsOnFloor || Puppet.LB["FootFront"].IsOnFloor)
                    {
                        MaxPayne(false);
                        Puppet.RB2["MiddleBody"].AddTorque(50f * -facing * Puppet.TotalWeight);
                        Puppet.RB2["Head"].AddForce(Vector2.up * 10 * Puppet.TotalWeight);
                        Puppet.RB2["Foot"].AddForce(Vector2.down * 5 * Puppet.TotalWeight);
                        Puppet.RB2["FootFront"].AddForce(Vector2.down * 5 * Puppet.TotalWeight);
                        state = DiveState.ready;
                    }




                }
                //if (state == DiveState.diving)
                //{
                //    Puppet.RB2["LowerBody"].AddForce(xdirection * (2.1f * Puppet.TotalWeight));
                //    Puppet.RB2["MiddleBody"].AddTorque(facing * (5 * Puppet.TotalWeight));

                //    if (frame > 5)
                //    {
                //        Puppet.PBO.OverridePoseIndex = -1;
                //        frame = 0;
                //        state = DiveState.landed;
                //    }

                //}

                //if (state == DiveState.landed) { 
                    
                //    if (Puppet.LB["Foot"].IsOnFloor || frame > 530)
                //    {
                //        ModAPI.Notify("Foot on floor");
                //        Puppet.RB2["MiddleBody"].AddTorque(-facing * (20 * Puppet.TotalWeight));
                //        Puppet.RB2["UpperBody"].AddRelativeForce(Vector2.up * 10 * Puppet.TotalWeight);

                //        ModAPI.Notify("landed");
                //        Puppet.PBO.OverridePoseIndex = -1;

                //        state = DiveState.ready;

                //        MaxPayne(false);

                //    }
                //}
            }
        }

        public enum ProneState
        {
            ready,
            start,
            drop,
            final,
        }

        public enum DiveState
        {
            ready,
            start,
            predive,
            diving,
            landed,
        };

    }

}