﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.Internal;
using UnityEngine.EventSystems;

namespace JTShtuff
{ 
    public class ChaseCam : MonoBehaviour
    {
        public static PhysicalBehaviour GPB         { get; set; }
        public static PhysicalBehaviour ChaseTarget { get; set; }
        public static PhysicalBehaviour LastTarget  { get; set; }
        public static GameObject HiddenFocus        { get; set; }
        public static int TargetHash                { get; set; } = 0;
        public static int LastHash                  { get; set; } = 0;
        public float Yoffset                        { get; set; } = 0;
        public float maxMoveSpeed                   { get; set; } = 25;
        public float smoothTime                     { get; set; } = 0.4f;
        public float minDistance                    { get; set; } = 2f;
        public float qTime                          { get; set; } = 0;
        public float sensitivity                    { get; set; } = 2f;
        public bool isChasing                       { get; set; } = false;

        public bool qChaseActive   = false;
        public bool mouseDown      = false;
        public bool targetClicked  = false;
        
        Vector2 currentVelocity;
        
        
        public void Start()
        {
            CreateChaseObject();
        }

        public void Chase(PhysicalBehaviour ctarget, int tHash)
        {
            if (!HiddenFocus || HiddenFocus.name != "ChaseCam") CreateChaseObject();

            ChaseTarget = ctarget;
            TargetHash  = tHash;

            JTUtil.Notify("Chase Cam Targeting: " + ChaseTarget.name, JTUtil.VerboseLevels.Full);

            HiddenFocus.transform.position = Global.CameraPosition;
            Yoffset                        = HiddenFocus.transform.position.y;

            Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();
            Global.main.CameraControlBehaviour.CurrentlyFollowing.Add(GPB);
            
            isChasing = true;
        }

        public void CreateChaseObject()
        {
            //  Create a hidden chase-cam object that the camera follows
            Texture2D tex          = new Texture2D(0, 0);
            Sprite mySprite        = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.1f, 0.1f), 0.1f);
            HiddenFocus            = ModAPI.CreatePhysicalObject("ChaseCam", mySprite);

            // Initialize chase-cam object
            HiddenFocus.transform.position = Global.CameraPosition;

            GPB = HiddenFocus.GetOrAddComponent<PhysicalBehaviour>();

            GPB.MakeWeightless();
            GPB.Selectable = false;
            GPB.Deletable  = false;

            LayerSerialisationBehaviour LSB = HiddenFocus.GetOrAddComponent<LayerSerialisationBehaviour>();
            LSB.Layer = 10;

            PuppetMaster.ChaseCam = this;

        }

        public void QuickChase(Thing thing, int seconds)
        {
            LastTarget  = ChaseTarget;
            LastHash    = TargetHash;

            ChaseTarget = thing.P;
            TargetHash  = thing.Hash;

            HiddenFocus.transform.position = Global.CameraPosition;
            Yoffset     = HiddenFocus.transform.position.y;

            Global.main.CameraControlBehaviour.CurrentlyFollowing.Clear();
            Global.main.CameraControlBehaviour.CurrentlyFollowing.Add(GPB);

            isChasing    = true;
            qChaseActive = true;
            qTime        = Time.time + seconds;
        }

        public void StopChase(int tHash)
        {
            if (tHash == TargetHash)
            {
                ChaseTarget = null;
                TargetHash  = 0;
                isChasing   = false;
            }
        }

        public void LateUpdate()
        {
            if (!isChasing) return;

            if (KB.MouseDown)
            {
                if (!mouseDown)
                {
                    mouseDown = true;
                    PersonBehaviour PB;

                    //  If user is dragging chaseTarget then dont follow
                    for (int i = SelectionController.Main.SelectedObjects.Count; --i >= 0;)
                    {
                        if (PB = SelectionController.Main.SelectedObjects[i].GetComponentInParent<PersonBehaviour>())
                        {
                            if (PB.GetHashCode() == PuppetMaster.ActivePuppet)
                            {
                                targetClicked = true;
                                break;
                            }
                        }
                    }
                }

                if (targetClicked) return;

            } else
            {
                mouseDown = targetClicked = false;
            }

            Vector2 chasePosition = ChaseTarget.transform.position;

            // Check that the distnace between target and hiddenobj is enough to care
            if (Mathf.Abs(Vector2.Distance((Vector2)HiddenFocus.transform.position, chasePosition)) < sensitivity) return;

            // Offsets the target position so that the object keeps its distance.
            chasePosition += ((Vector2)HiddenFocus.transform.position - chasePosition).normalized * minDistance;

            if (chasePosition.y < Yoffset) chasePosition.y = Yoffset;

            HiddenFocus.transform.position = Vector2.SmoothDamp(HiddenFocus.transform.position, chasePosition, ref currentVelocity, smoothTime, maxMoveSpeed);
        }

        public void Update()
        {
            if (qChaseActive && Time.time > qTime)
            {
                qChaseActive = false;
                qTime        = 0;

                Chase(LastTarget, LastHash);
            }
        }
    }
}
