﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Unity;
using UnityEngine;
using UnityEngine.EventSystems;
using UnityEngine.Events;

namespace JTShtuff
{
    public static class KB
    {
        public const float DoubleTap      = 0.5f;
        public static bool UsingCustom    { get; set; } = false;
        public static bool PClick         { get; set; } = false;
        public static bool Up             { get; set; } = false;
        public static bool Down           { get; set; } = false;
        public static bool Left           { get; set; } = false;
        public static bool Right          { get; set; } = false;
        public static bool Action         { get; set; } = false;
        public static bool ActionHeld     { get; set; } = false;
        public static bool Throw          { get; set; } = false;
        public static bool Aim            { get; set; } = false;
        public static bool Shift          { get; set; } = false;
        public static bool Control        { get; set; } = false;
        public static bool Alt            { get; set; } = false;
        public static bool Modifier       { get; set; } = false;
        public static bool MouseDown      { get; set; } = false;
        public static bool AnyKey         { get; set; } = false;
        public static bool Emote          { get; set; } = false;
        public static bool Inventory      { get; set; } = false;
        public static bool isNumDisabled  { get; set; } = false;
        
        public static class KeyTimes
        {
            public static float Left   { get; set; } = 0f;
            public static float Right  { get; set; } = 0f;
            public static float Up     { get; set; } = 0f; 
            public static float Down   { get; set; } = 0f;
        }
        public static class KeyCombos
        {
            public static bool DoubleRight { get; set; } = false;
            public static bool DoubleLeft  { get; set; } = false;
            public static bool DoubleUp    { get; set; } = false;
            public static bool DoubleDown  { get; set; } = false;
        }


        public static Dictionary<string, InputAction> DefaultBindings   = new Dictionary<string, InputAction>();
        public static Dictionary<string, InputAction> PuppetBindings    = new Dictionary<string, InputAction>();
        public static Dictionary<string, KeyCode> DisabledKeys          = new Dictionary<string, KeyCode>();
        public static List<KeyCode> PuppetKeys                          = new List<KeyCode>();
        

        public static void CheckKeys()
        {
            PClick = Up = Down = Left = Right = Action = ActionHeld = Aim = Shift = Control = Alt = Modifier = MouseDown = Emote = Inventory = false;

            if (Input.GetKey(KeyCode.None)) return;

            Shift    = (Input.GetKey(KeyCode.LeftShift)   || Input.GetKey(KeyCode.RightShift));
            Alt      = (Input.GetKey(KeyCode.LeftAlt)     || Input.GetKey(KeyCode.RightAlt));
            Control  = (Input.GetKey(KeyCode.LeftControl) || Input.GetKey(KeyCode.RightControl));
            Modifier = Shift || Alt || Control;

            PClick     = Modifier && Input.GetMouseButtonUp(0);
            MouseDown  = Input.GetMouseButton(0);
            Up         = InputSystem.Held("PM-MoveUp");
            Down       = InputSystem.Held("PM-MoveDown");
            Left       = InputSystem.Held("PM-MoveLeft");
            Right      = InputSystem.Held("PM-MoveRight");
            Throw      = InputSystem.Held("PM-Throw");
            Emote      = InputSystem.Held("PM-Emote");
            ActionHeld = InputSystem.Held("PM-Action");
            Inventory  = InputSystem.Held("PM-Inventory");
            Action     = InputSystem.Down("PM-Action");
            Aim        = InputSystem.Down("PM-Aim");

            AnyKey = Up || Down || Left || Right || Action || Throw || Aim || Emote || Inventory;

            if (InputSystem.Down("PM-MoveUp") )
            {
                KeyCombos.DoubleUp     = (Time.time - KeyTimes.Up < DoubleTap);
                KeyTimes.Up            = Time.time;
            }

            if (InputSystem.Down("PM-MoveDown"))
            {
                KeyCombos.DoubleDown   = (Time.time - KeyTimes.Down < DoubleTap);
                KeyTimes.Down          = Time.time;
            }

            if (InputSystem.Down("PM-MoveLeft"))
            {
                KeyCombos.DoubleLeft   = (Time.time - KeyTimes.Left < DoubleTap);
                KeyTimes.Left          = Time.time;
            }

            if (InputSystem.Down("PM-MoveRight"))
            {
                KeyCombos.DoubleRight  = (Time.time - KeyTimes.Right < DoubleTap);
                KeyTimes.Right         = Time.time;
            }
        }

        public static void InitControls()
        {
            string[] ControlSchematic =
            {
                "MoveLeft    > Move Puppet Left        > A",
                "MoveRight   > Move Puppet Right       > D",
                "MoveUp      > Move Up / Jump          > W",
                "MoveDown    > Move Down / Duck        > S",
                "Action      > Use / Attack            > K",
                "Throw       > Throw Item              > T",
                "Aim         > Aim Weapon              > R",
                "Inventory   > Inventory               > BackQuote",
                "Emote       > Emote (Followed by #)   > Semicolon",
                "TogglePM    > Toggle Puppet Master    > F7",
            };
            
            string inputKey;
            InputAction action;

            foreach (string controlMapping in ControlSchematic)
            {
                string[] KeyParts = controlMapping.Split('>');

                inputKey = "PM-" + KeyParts[0].Trim();

                if (!InputSystem.Has(inputKey))
                {
                    ModAPI.RegisterInput("[PM] " + KeyParts[1].Trim(), inputKey, (KeyCode)Enum.Parse(typeof(KeyCode), KeyParts[2].Trim()));
                }
                else
                {
                    action = InputSystem.Actions[inputKey];
                    ModAPI.RegisterInput("[PM] " + KeyParts[1].Trim(), inputKey, action.Key);

                    InputAction actionNew  = InputSystem.Actions[inputKey];
                    actionNew.SecondaryKey = action.SecondaryKey;
                }
            }

            //  Save the Default Bindings
            foreach (KeyValuePair<string, InputAction> xaction in InputSystem.Actions)
            {
                if (xaction.Key.Contains("PM") && !xaction.Key.Contains("TogglePM"))
                {
                    PuppetBindings.Add(xaction.Key, xaction.Value);

                    PuppetKeys.Add(xaction.Value.Key);
                }

                else DefaultBindings.Add(xaction.Key, xaction.Value);
            }
        }


        public static void SwapBindings(bool useCustom = false)
        {
            JTUtil.Notify("<color=red>Swapping Bindings to:</color> " + (useCustom ? "CUSTOM" : "DEFAULT"), JTUtil.VerboseLevels.Full); 

            UsingCustom = useCustom;

            foreach (KeyValuePair<string, InputAction> keyset in PuppetBindings)
            {
                if (useCustom) if (!InputSystem.Has(keyset.Key)) InputSystem.Actions.Add(keyset.Key, keyset.Value);

                else if (!InputSystem.Has(keyset.Key)) InputSystem.Actions.Remove(keyset.Key);
            }

            foreach (KeyValuePair<string, InputAction> keyset in DefaultBindings)
            {
                if (useCustom)
                {
                    if (PuppetKeys.Contains(keyset.Value.Key))
                    {
                        if (InputSystem.Has(keyset.Key)) InputSystem.Actions[keyset.Key].Key = KeyCode.Clear;
                    }
                }

                else
                {
                    if (!InputSystem.Has(keyset.Key)) InputSystem.Actions.Add(keyset.Key, keyset.Value);
             
                    else InputSystem.Actions[keyset.Key].Key = keyset.Value.Key;
                }
            }
        }


        public static void DisableNumberKeys()
        {
            if (isNumDisabled) return;

            isNumDisabled = true;

            DisabledKeys.Clear();

            foreach (KeyValuePair<string, InputAction> actionKey in InputSystem.Actions)
            {
                if ( (int)actionKey.Value.Key >= 48 && (int)actionKey.Value.Key <= 57 )
                {
                    DisabledKeys.Add(actionKey.Key, actionKey.Value.Key);
                    InputSystem.Actions[actionKey.Key].Key = KeyCode.Clear;
                }
            }
        }

        public static void EnableNumberKeys()
        {
            if (!isNumDisabled) return;

            isNumDisabled = false;

            foreach (KeyValuePair<string, KeyCode> keyset in DisabledKeys)
            {
                InputSystem.Actions[keyset.Key].Key = keyset.Value;
            }
        }

        public static int CheckNumberKey()
        {
            for (int i = 48; i <= 57; i++)
            {
                if (Input.GetKeyDown((KeyCode)i)) return i - 48;
            }

            return -1;
        }
    }
}
