﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using Newtonsoft.Json;
using System.Runtime.Serialization.Formatters;

namespace JTShtuff
{
    public class CustomPoses
    {
        public static Dictionary<string, PoseSet> Poses = new Dictionary<string, PoseSet>();

        public class PoseSet
        {
            public Dictionary<string, string> LimbAngles { get; set; }
            public bool Animated                         { get; set; } = false;
            public float AnimationSpeedMultiplier        { get; set; } = 1f;
            public float DragInfluence                   { get; set; } = 1.0f;
            public float RandomInfluence                 { get; set; } = 0f;
            public float RandomSpeed                     { get; set; } = 0f;
            public float Rigidity                        { get; set; } = 5f;
            public bool ShouldStumble                    { get; set; } = true;
            public bool ShouldStandUpright               { get; set; } = true;
            public PoseState State                       { get; set; } = 0;
            public float TimeOffset                      { get; set; } = 0f;
            public float UprightForceMultiplier          { get; set; } = 1.0f;
        }


        public static Dictionary<string, int> Assign(PersonBehaviour Peep)
        {
            Dictionary<string, int> poseList = new Dictionary<string, int>();

            foreach (string poseName in Poses.Keys)
            {
                poseList.Add(poseName, GetCustomPose(Peep, poseName));
            }

            return poseList;
        }

        // - - - - - - - - - - - - - - - - - - -
        //
        //  FN: GET CUSTOM POSE
        //  Checks if peep has a custom pose and creates it if necessary
        //
        public static int GetCustomPose(PersonBehaviour Peep, string PoseName)
        {
            if (!Poses.ContainsKey(PoseName))
            {
                JTUtil.Notify( "<color=red>Missing Custom Pose:</color> " + PoseName, JTUtil.VerboseLevels.Full);
                return 0;
            }
            
            for (int pnum = Peep.Poses.Count; --pnum >= 0;)
                if (Peep.Poses[pnum].Name == PoseName) return pnum;

            RagdollPose customPose = new RagdollPose()
            {
                Name                     = PoseName,
                Angles                   = new List<RagdollPose.LimbPose>(),
                ShouldStandUpright       = Poses[PoseName].ShouldStandUpright,
                AnimationSpeedMultiplier = Poses[PoseName].AnimationSpeedMultiplier,
                UprightForceMultiplier   = Poses[PoseName].UprightForceMultiplier,
                DragInfluence            = Poses[PoseName].DragInfluence,
                Rigidity                 = Poses[PoseName].Rigidity,
                State                    = PoseState.Rest,

            };

            string limbAngleIn;

            foreach (LimbBehaviour limb in Peep.Limbs)
            {
                if (Poses[PoseName].LimbAngles.TryGetValue(limb.name, out limbAngleIn) )
                {
                    float startAngle = 0.0f;
                    float endAngle   = 0.0f;

                    if (limbAngleIn.Contains(":"))
                    {
                        // Angle range
                        string[] angleRange = limbAngleIn.Split(':');
                        float.TryParse(angleRange[0], out startAngle);
                        float.TryParse(angleRange[1], out endAngle);
                    }
                    else
                    {
                        // single angle
                        float.TryParse(limbAngleIn, out startAngle);

                        endAngle = startAngle;

                    }

                    customPose.Angles.Add(new RagdollPose.LimbPose(limb, startAngle)
                    {
                        Animated   = Poses[PoseName].Animated,
                        StartAngle = startAngle,
                        EndAngle   = endAngle + 5f
                    });
                }
            }

            customPose.ConstructDictionary();

            Peep.Poses.Add(customPose);

            int poseId = Peep.Poses.Count - 1;
            
            Peep.SetPose(Peep.Poses[poseId]);
            
            return poseId;
        }



        // - - - - - - - - - - - - - - - - - - -
        //
        //  FN: CHANGE POSE
        //
        public static void Pose(string poseName, PersonBehaviour Peep)
        {
            int poseId = GetCustomPose(Peep, poseName);

            Peep.OverridePoseIndex = poseId;
        }

        public static void InitCustomPoses()
        {
            string JTCustomPoses = @"
            {
                'JT_Duck': {
                    'LimbAngles':  {
                        'Foot':                 '-24.21083', 
                        'UpperLeg':             '-0.03494752', 
                        'LowerLeg':             '106.0683', 
                        'LowerLegFront':        '78.07973', 
                        'UpperLegFront':        '-75.54669:', 
                        'FootFront':            '0.4032795', 
                        'Head':                 '-0.0002471675', 
                        'LowerBody':            '-0.0002010637',
                        'UpperArm':             '0.1',
                        'UpperArmFront':        '0.1'
                    }, 
                    'ShouldStandUpright': true, 
                    'Animated': false, 
                    'AnimationSpeedMultiplier': 1.0
                },

                'JT_FrontFlip': {
                    'LimbAngles': {
                        'LowerArm':             '-95.09151', 
                        'LowerArmFront':        '-85.78412', 
                        'UpperArm':             '-376.2525', 
                        'Head':                 '6.57841', 
                        'MiddleBody':           '0', 
                        'UpperArmFront':        '-369.4041', 
                        'UpperBody':            '0.4076479', 
                        'LowerBody':            '-29.47088', 
                        'UpperLegFront':        '-111.5341', 
                        'UpperLeg':             '-112.034', 
                        'LowerLeg':             '133.9088', 
                        'LowerLegFront':        '134.5014', 
                        'Foot':                 '-14.32194', 
                        'FootFront':            '-14.81616'
                    },
                    'ShouldStandUpright': true, 
                    'AnimationSpeedMultiplier': 0.5,
                    'Animated': false,
                    'UprightForceMultiplier': 2,
                    'Rigidity': 1.3
                },

                'JT_Dive_1': {
                    'LimbAngles': {
                        'LowerLegFront':        '92.51105', 
                        'LowerLeg':             '97.91881', 
                        'MiddleBody':           '0', 
                        'FootFront':            '-0.0001707547', 
                        'Foot':                 '-0.0001878302', 
                        'UpperBody':            '-18.16421', 
                        'UpperArm':             '-35.69885', 
                        'Head':                 '15.33757', 
                        'UpperArmFront':        '-39.43206', 
                        'LowerArm':             '-51.0381', 
                        'UpperLegFront':        '-111.6845', 
                        'LowerArmFront':        '-51.81177', 
                        'LowerBody':            '-17.23651', 
                        'UpperLeg':             '-112.5029'
                    },
                    'ShouldStandUpright': false, 
                    'AnimationSpeedMultiplier': 10.5,
                    'Animated': false,
                    'UprightForceMultiplier': 1.1,
                    'Rigidity': 4.1
                },

                'JT_Dive_2': {
                    'LimbAngles': {
                        'MiddleBody':           '0', 
                        'UpperArmFront':        '-149.6484', 
                        'UpperBody':            '-13.5377', 
                        'Head':                 '-3.796226', 
                        'LowerArmFront':        '-9.33497', 
                        'LowerArm':             '0.02141264', 
                        'UpperArm':             '-142.3777', 
                        'LowerLegFront':        '54.19135', 
                        'LowerLeg':             '45.83427', 
                        'UpperLegFront':        '-50.31013', 
                        'UpperLeg':             '-43.46162', 
                        'FootFront':            '-31.9845', 
                        'Foot':                 '-32.13535', 
                        'LowerBody':            '-1.615463'
                    },
                    'ShouldStandUpright': true, 
                    'AnimationSpeedMultiplier': 10.5,
                    'Animated': false,
                    'UprightForceMultiplier': 5,
                    'Rigidity': 3.5
                },

                'JT_Prone': {
                    'LimbAngles': {
                        'Head': '-6.500257', 
                        'LowerBody':            '17.68937', 
                        'MiddleBody':           '0', 
                        'UpperArmFront':        '-18.54307', 
                        'UpperArm':             '-16.35537', 
                        'LowerLegFront':        '9.899184', 
                        'LowerLeg':             '11.33566', 
                        'Foot':                 '0.02769642', 
                        'FootFront':            '0.04737419', 
                        'UpperLegFront':        '16.45826', 
                        'UpperLeg':             '15.87339', 
                        'LowerArmFront':        '-103.8956', 
                        'UpperBody':            '-17.44694', 
                        'LowerArm':             '-102.0114'
                    },
                    'ShouldStandUpright': false, 
                    'AnimationSpeedMultiplier': 10.5,
                    'Animated': false,
                    'UprightForceMultiplier': 1,
                    'Rigidity': 2.2,
                    'State': 8
                },

                'JT_Waving': {
                    'LimbAngles': {
                        'LowerArm':      '-0.08787935', 
                        'LowerBody':     '0.1363826', 
                        'MiddleBody':    '0', 
                        'UpperLegFront': '-0.2402655', 
                        'UpperLeg':      '0.6425738', 
                        'LowerLegFront': '-0.1414893', 
                        'LowerLeg':      '0.06455169', 
                        'UpperArm':      '0.1430857', 
                        'Head':          '0.01760935', 
                        'UpperBody':     '0.01076555', 
                        'LowerArmFront': '-64.94197', 
                        'UpperArmFront': '-87.36889', 
                        'Foot':          '-0.5595296', 
                        'FootFront':     '0.1156012'
                    },
                    'ShouldStandUpright': true, 
                    'Animated': false,
                    'UprightForceMultiplier': 0.1,
                    'Rigidity': 5.2
                },

                'JT_Throw_1': {
                    'LimbAngles': {
                        'LowerArm':      '0.1230614', 
                        'UpperArm':      '-0.117004', 
                        'LowerLeg':      '-0.001100474', 
                        'LowerLegFront': '0.001414623', 
                        'UpperArmFront': '174.287', 
                        'LowerArmFront': '-86.93764', 
                        'Foot':          '0.1132696', 
                        'FootFront':     '0.1011531', 
                        'UpperLeg':      '-0.007332622', 
                        'UpperLegFront': '0.00374265', 
                        'MiddleBody':    '0', 
                        'UpperBody':     '-0.0001217294', 
                        'Head':          '0.0004742446', 
                        'LowerBody':     '-0.000948049'
                    },
                    'ShouldStandUpright': true, 
                    'Animated': false,
                    'UprightForceMultiplier': 0.1,
                    'Rigidity': 5.2
                },

                'JT_Throw_2': {
                    'LimbAngles': {
                        'LowerLeg':      '33.29119', 
                        'Foot':          '-6.515359', 
                        'UpperLeg':      '-31.82211', 
                        'UpperLegFront': '13.24422', 
                        'LowerArm':      '-0.01831088', 
                        'LowerLegFront': '-0.02533829', 
                        'FootFront':     '-5.199605', 
                        'LowerBody':     '-0.02204017', 
                        'MiddleBody':    '0', 
                        'UpperBody':     '-0.002199321', 
                        'Head':          '-0.00341168', 
                        'UpperArmFront': '237.4726', 
                        'UpperArm':      '0.01295345', 
                        'LowerArmFront': '-33.08913'
                    },
                    'ShouldStandUpright': true, 
                    'Animated': false,
                    'UprightForceMultiplier': 0.1,
                    'Rigidity': 5.2
                },

                'JT_Throw_x': {
                    'LimbAngles': {
                        'UpperArmFront': '53.87996', 
                        'UpperArm':      '0.1090901', 
                        'UpperBody':     '-0.003569094', 
                        'Head':          '-0.07377298', 
                        'LowerLegFront': '-0.01318056', 
                        'FootFront':     '-0.8645177', 
                        'LowerArmFront': '-36.86349', 
                        'LowerArm':      '-0.4033567', 
                        'UpperLegFront': '3.002661', 
                        'UpperLeg':      '-0.003749454', 
                        'LowerBody':     '-0.05592484', 
                        'MiddleBody':    '0', 
                        'LowerLeg':      '-0.007670409', 
                        'Foot':          '0.9779164'
                    },
                    'ShouldStandUpright': true, 
                    'Animated': false,
                    'UprightForceMultiplier': 0.1,
                    'Rigidity': 5.2
                },

                'JT_Victory': {
                    'LimbAngles': {
                        'LowerArmFront': '-42.56367', 
                        'Head':          '0.002119066', 
                        'LowerArm':      '-56.61966', 
                        'LowerBody':     '-0.005534801', 
                        'LowerLeg':      '-0.0006021239', 
                        'LowerLegFront': '0.002093026', 
                        'FootFront':     '-1.31155', 
                        'UpperLeg':      '-0.002053326', 
                        'UpperLegFront': '-0.001093257', 
                        'Foot':          '-1.019251', 
                        'UpperArm':      '205.2887', 
                        'MiddleBody':    '0', 
                        'UpperArmFront': '-123.5136', 
                        'UpperBody':     '-0.004975366'
                    },
                    'ShouldStandUpright': true, 
                    'Animated': false,
                    'UprightForceMultiplier': 1.1,
                    'Rigidity': 5.2       
                }
            }";
            
            Poses = JsonConvert.DeserializeObject<Dictionary<string, PoseSet>>(JTCustomPoses);

        }

        public static void LittleMovement(Rigidbody2D rb, float minVal=50f, float maxVal=60f, bool bothDirections=true)
        {
            float force = GetRandomNumber(minVal, maxVal);

            int direction = 1;
            if (bothDirections) {
                
                direction = getrandom.Next(0,1);

            }

            if (direction <= 0) force *= -1;
            
            rb.MoveRotation(force);
        }

        public static readonly System.Random getrandom = new System.Random();
        public static float GetRandomNumber(float min, float max)
        {
            int imin = (int)Math.Round(min * 100);
            int imax = (int)Math.Round(max * 100);
            
            lock (getrandom) // synchronize
            {
                return (float)Mathf.Round(getrandom.Next(imin, imax) / 100);
            }
        }
    }
}
