﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using JTShtuff;

namespace JTShtuff
{
    public class InventoryBehaviour : MonoBehaviour
    {
        public PersonBehaviour PBO              { get; set; }
        public PuppetBehaviour Puppet           { get; set; }
        public Thing Holding                    { get; set; }
        public int CurrentSpot                  { get; set; } = -1;
        public Dictionary<int, Thing> Inventory { get; set; } = new Dictionary<int, Thing>();
        public Thing Clone                      { get; set; }
        public bool Enabled                     { get; set; } = true;

        private bool InventoryTriggered         = false;
        private bool InventoryDone              = false;
        private bool doEquipClone               = false;
        private bool doPutAway                  = false;
        private bool doFreeze                   = false;
        private bool doUnfreeze                 = false;
        private bool TriggerPickup              = false;


        public void Update()
        {
            if (!Enabled || !Puppet.isActive || !PuppetMaster.Enabled) return;

            if (TriggerPickup) { 
                
                TriggerPickup = false;  
                
                Puppet.DoPickup(Clone.P); 
                
            }

            if (KB.Inventory)
            {
                if (!InventoryTriggered) InitInventory();
                else
                {
                    int spot = KB.CheckNumberKey();

                    if (spot > -1 && CurrentSpot != spot) CurrentSpot = spot;

                    if (KB.MouseDown)
                    {
                        Thing thing = new Thing(JTUtil.GetClickedItem());

                        if (thing != null) StoreItem(thing);
                    }
                }
            }

            if (InventoryTriggered && !KB.Inventory) ExecuteInventory();

        }

        public void LateUpdate()
        {
            if (doFreeze)
            {
                foreach (KeyValuePair<int,Thing> pair in Inventory)
                {
                    FreezeStackController.RequestFreeze(pair.Value.R);
                }

                doFreeze = false;

            } else if (doUnfreeze)
            {
                FreezeStackController.RequestUnfreeze(Clone.R);

                doUnfreeze = false;
            }
        }


        public void Awake()
        {
            PBO         = GetComponent<PersonBehaviour>();
            Puppet      = GetComponent<PuppetBehaviour>();
        }
        


        public void InitInventory()
        {
            KB.DisableNumberKeys(); 

            InventoryTriggered  = true;
            InventoryDone       = false;
            CurrentSpot         = -1;

            //  Determine if puppet is holding something
            Holding = Thing.HeldThing(PBO);
            Puppet.AimingStop();
            Puppet.isAiming = false;
        }



        public void ExecuteInventory()
        {
            Puppet.AimingStop();

            InventoryTriggered  = false;

            KB.EnableNumberKeys();

            if (InventoryDone) return;

            InventoryDone       = true;

            if ((Holding == null || Holding.Name == null) && CurrentSpot == -1)
            {
                 JTUtil.Notify("<color=orange>Hold </color>[ I ]<color=orange> and press </color>#[ 1-9 ]<color=orange> to store/retrieve Items</color>");
                 return;
            }

            if (Holding != null) StoreItem(Holding, -1);

            if (CurrentSpot > -1) SelectItem(CurrentSpot);
        }



        public bool StoreItem(Thing thing, int spot = -1)
        {
            if (Puppet.isAiming) Puppet.AimingStop();
            
            //  Check if we already have blueprint of this item
            int curpos = LocateItem(thing);

            if (curpos > -1)
            {
                //  We already have the item blueprint, so destroy this one
                thing.Destroy();

                JTUtil.Notify(thing.Name + " <color=orange>was put away</color>", JTUtil.VerboseLevels.Full);

                return true;
            }


            if(spot == -1) spot = GetOpenInventorySpot();
            
            if (spot == -1)
            {
                JTUtil.Notify("<color=orange>Inventory Full</color>");
                return false;
            }

            CleanItem(thing);

            
            Inventory.Add(spot, thing);

            InventoryDone = true;


            JTUtil.Notify(thing.Name + " <color=orange>stored in position #:</color> " + spot);

            return true;

        }

        

        public int GetOpenInventorySpot()
        {
            int spot = -1;

            for (int i = 0; ++i <= 10;)
            {
                if (i==10) i = 0;

                if (!Inventory.ContainsKey(i)) return i;
            }

            return spot;
        }



        public int LocateItem(Thing thing)
        {
            foreach (KeyValuePair<int, Thing> pair in Inventory)
            {
                if (pair.Value.Hash == thing.Hash || pair.Value.Name == thing.Name) return pair.Key;
            }

            return -1;
        }




        public void CleanItem(Thing thing)
        {
            thing.BreakConnections();

            thing.G.layer = 10;

            thing.G.transform.position = (new Vector2(-100, -100));

            thing.R.position = (new Vector2(1, 1));
            
            thing.P.MakeWeightful();

            doFreeze = true;

        }



        public bool SelectItem(int spotChk)
        {
            if (!Inventory.TryGetValue(spotChk, out Thing thing)) return false;
            
            GameObject instance = UnityEngine.Object.Instantiate(thing.G, thing.G.transform.position, Quaternion.identity);
            
            instance.name  = thing.Name;
            Clone          = new Thing(instance);

            Clone.R.bodyType = RigidbodyType2D.Dynamic;
            Clone.G.layer    = 9;

            PuppetMaster.LastClickedItem = Clone.P;

            TriggerPickup = true;
            InventoryDone = true;

            Puppet.ItemFH = Clone.P;
            
            return true;

        }
       
    }

}