﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;

namespace JTShtuff
{
    public class PuppetBehaviour : MonoBehaviour
    {
        const float DoubleTap = 0.5f;
        public int MovementActive      { get; set; } = 0;
        public bool isActive           { get; set; } = false;
        public bool isHolding          { get; set; } = false;
        public bool isWalking          { get; set; } = false;
        public bool isCrouching        { get; set; } = false;
        public bool isProne            { get; set; } = false;
        public bool isJumping          { get; set; } = false;
        public bool isAiming           { get; set; } = false;
        public bool isFiring           { get; set; } = false;
        public bool isThrowing         { get; set; } = false;
        public bool doneThrowing       { get; set; } = true;
        public bool jumpLocked         { get; set; } = false;
        public bool isEmote            { get; set; } = false;
        public bool canShoot           { get; set; } = false;
        public bool useAuto            { get; set; } = false;
        public bool hasJumped          { get; set; } = false;
        public bool facingLeft         { get; set; } = false;
        public int EmoteId             { get; set; } = -1;
        public float JumpStrength      { get; set; } = 0;
        public float DisableMoves      { get; set; } = 0;
        public float TotalWeight       { get; set; } = 0;
        public float ThrowWeight       { get; set; } = 0; 
        public float ThrowForce        { get; set; } = 0; 
        public bool enableCameraFollow { get; set; } = true;
        public float TimedCall         { get; set; } = 0;
        public float TimeAnimation     { get; set; } = 0;
        

        public Emotes Emote            { get; set; } = 0;

        public Dictionary<string, int> PuppetPose { get; set; } = new Dictionary<string, int>();

        public PersonBehaviour PBO                  { get; set; } = null;
        public InventoryBehaviour Inventory         { get; set; } = null;
        public JTActionsBehaviour Actions           { get; set; } = null;
        public PhysicalBehaviour ItemFH             { get; set; } = null;
        public PhysicalBehaviour LastClicked        { get; set; } = null;
        public Thing Holding                        { get; set; } = null;
        public GripBehaviour GripF                  { get; set; } = null;
        public GripBehaviour GripB                  { get; set; } = null;
        public Dictionary<string, Rigidbody2D> RB2  { get; set; } = new Dictionary<string, Rigidbody2D>();
        public Dictionary<string, LimbBehaviour> LB { get; set; } = new Dictionary<string, LimbBehaviour>();
        
        public Dictionary<string, LimbSnapshot> LimbSnapshots = new Dictionary<string, LimbSnapshot>();
        public struct LimbSnapshot
        {
            public float Health;
            public float Numbness;
            public float BreakingThreshold;
            public float Vitality;
            public float RegenerationSpeed;
            public float BaseStrength;
            public bool Broken;
            public bool Frozen;
            public ushort BruiseCount;

        }

        public enum Emotes {
            None,
            Waving,
            Victory,
            Prone,
            Dive1,
            Dive2,
        };

        public float TimeJump          = 0f;
        public float cameraSpeed       = 0.1f;

        public Vector2 aimVelocity, aimVelocityRev;
        public Vector3 cameraOffset;

        public List<PoseState> DisablingPoses = new List<PoseState>() {
            PoseState.Flailing,
            PoseState.Protective,
            PoseState.WrithingInPain 
        };
        
        public bool isPuppetReady()      => (bool)(PuppetMaster.Enabled 
                                                   && isActive 
                                                   && PBO.isActiveAndEnabled 
                                                   && PBO.IsAlive() 
                                                   && PBO.Consciousness >= 1 
                                                   && PBO.ShockLevel < 0.3f );
        public bool PuppetFlip()         => (bool)(!DisabledMoves() 
                                                   && !DisablingPoses.Contains(PBO.ActivePose.State) 
                                                   && isUpright()
                                                   && !(KB.Left && KB.Right)
                                                   && !isJumping
                                                   && JTUtil.FlipPerson(PBO)); 
        public bool isUpright()          => (bool)(!LB["LowerBody"].IsOnFloor
                                                   && !LB["UpperBody"].IsOnFloor
                                                   && !LB["Head"].IsOnFloor
                                                   && LB["FootFront"].IsOnFloor
                                                   && LB["Foot"].IsOnFloor);
        public bool isFlipped()          => (bool)(PBO.transform.localScale.x < 0.0f);
        public bool DisabledMoves()      => (bool)(DisableMoves > Time.time);


        public void Start()
        {

            //  Check if this puppet is the current active puppet
            if (PBO.GetHashCode() != PuppetMaster.ActivePuppet)
            {
                isActive = false; 
                Destroy(this);
                return;
            }

            Strengthen();

            JTUtil.Notify("Puppet set: <color=yellow>" + PBO.name + "</color>");

            PhysicalBehaviour PBX = PBO.GetComponentInChildren<PhysicalBehaviour>();
            
            if (PBX != null) PuppetMaster.ChaseCam.Chase(PBX, this.GetHashCode());

        }


        public void Awake()
        {
            TotalWeight       = 0f;
            PBO               = GetComponent<PersonBehaviour>();

            //  Cache Rigidbody Maps
            Rigidbody2D[] RBs = PBO.transform.GetComponentsInChildren<Rigidbody2D>();
            
            foreach (Rigidbody2D rb in RBs)
            {
                RB2.Add(rb.name, rb);

                TotalWeight += rb.mass;

                switch (rb.name)
                {
                    case "LowerArmFront":
                    case "LowerArm":
                        GripBehaviour grip = rb.name == "LowerArm" ? GripB : GripF;
                        if (rb.TryGetComponent<GripBehaviour>(out GripBehaviour GB)) grip = GB;
                        break;
                }
            }

            //  Cache LimbBheaviour Maps
            LimbBehaviour[] LBs = PBO.GetComponentsInChildren<LimbBehaviour>();

            foreach (LimbBehaviour limb in LBs) 
            {
                LB.Add(limb.name, limb);
            }

            if (GripF != null && GripF.CurrentlyHolding) ItemFH = GripF.CurrentlyHolding;

            if (PBO.OverridePoseIndex != (int)PoseState.Rest || PBO.OverridePoseIndex != (int)PoseState.Sitting)
            {
                PBO.OverridePoseIndex = -1;
            }

            //  Assign custom poses
            PuppetPose  = CustomPoses.Assign(PBO);


            //  Init Inventory Behaviour
            Inventory         = gameObject.GetOrAddComponent<InventoryBehaviour>();
            Inventory.Enabled = true;

            //  Init Actions Behaviour
            Actions           = gameObject.GetOrAddComponent<JTActionsBehaviour>();
            Actions.Puppet    = this;
        }


        private void FixedUpdate()
        {
            if (!isActive) return;
            

            if (isJumping && !KB.Up) DoJump(true);

            if (isAiming) { 
                if (!isPuppetReady() )
                {
                    isAiming = false;
                    return;
                }

                AimWeapon();
            }

        }

        public void Update()
        {
            if (!isActive) return;

            if (KB.AnyKey && !isPuppetReady()) return;

            if (!KB.ActionHeld && !KB.MouseDown) isFiring = false;


            if (isWalking)
            {
                if (KB.Left && !facingLeft && PuppetFlip())      facingLeft = true;
                else if (KB.Right && facingLeft && PuppetFlip()) facingLeft = false;
                else
                {
                    isWalking  = false;
                    
                    StopPerson();
                }
            }

            if (isCrouching)
            {
                
                if (!KB.Down)
                {
                    isCrouching = false;
                    
                    StopPerson();
                    
                }
                else
                {
                    if (KB.Left || KB.Right)
                    {
                        if (facingLeft != KB.Left && isPuppetReady()) {
                            
                            PuppetFlip();

                            facingLeft = isFlipped();

                        }
                    }
                }
            }


            if (!DisabledMoves()) CheckControls();
            

            if (KB.Emote) { 
                
                doEmote();
                
            
            } else if(isEmote)
            {
                isEmote = false;
                KB.EnableNumberKeys();
                if (EmoteId == -1) JTUtil.Notify("HOLD THIS KEY & CHOOSE EMOTE #", JTUtil.VerboseLevels.Minimal);
            }

            if ( KB.Aim ) 
            {
                //  ---
                //  Keys AIM
                //  ---
                isAiming = !isAiming;

                if (isAiming) {
                    
                    AimingInit();

                }
                else {
                    
                    AimingStop();
                }

            }

            if (KB.Throw)
            {
                //  ---
                //  Keys Throw
                //  ---
                if( !isThrowing ) DoThrow();
                else if(KB.ActionHeld || KB.MouseDown)
                {
                    Holding.G.SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                }

            } else if (isThrowing)
            {
                DoThrow(true);

            } else if (!doneThrowing && TimeAnimation < Time.time)
            {
                doneThrowing = true;
                PBO.OverridePoseIndex = -1;
                ThrowHeldItem(ThrowForce);
            }

             
             
            
        }


        private void CheckControls()
        {



            if (KB.Left || KB.Right)
            {
                //  ---
                //  Keys LEFT + RIGHT
                //  ---
                if ( PBO.IsTouchingFloor && !isWalking && !KB.Up && !isProne)
                {

                    if ((KB.KeyCombos.DoubleRight || KB.KeyCombos.DoubleLeft))
                    {
                        if (isFlipped() != KB.Left)
                        {
                            JTUtil.FlipPerson(PBO);
                            facingLeft = isFlipped();
                        }

                        KB.KeyCombos.DoubleRight = KB.KeyCombos.DoubleLeft = false;
                        
                        Actions.dive.Init();

                    }

                    else if (!isCrouching && PBO.IsTouchingFloor)
                    {
                        if (isFlipped() != KB.Left) PuppetFlip();

                        facingLeft = isFlipped();
                        PBO.OverridePoseIndex = (int)PoseState.Walking;
                        isWalking = true;
                    }
                }
            } 

            else if (KB.Down && !isCrouching && !isProne && PBO.IsTouchingFloor)
            {
                //  ---
                //  Keys DOWN
                //  ---
                if (KB.KeyCombos.DoubleDown)
                {
                    Actions.prone.Init();
                    KB.KeyCombos.DoubleDown = false;
                } 
                else { 

                    if (KB.Modifier) PBO.OverridePoseIndex = (int)PoseState.Sitting;
                    else DoCrouch();

                }

            }

            else if (KB.Up)
            {
                if (isProne)
                {
                    StopPerson();
                    isProne    = false;
                    jumpLocked = true;
                } else
                if (PBO.IsTouchingFloor && !isCrouching && !jumpLocked)
                {
                    if (!DisabledMoves()) DoJump();
                }
            } else if (jumpLocked) jumpLocked = false;

            if (KB.Up && PBO.IsTouchingFloor && !isCrouching)
            {
                //  ---
                //  Keys UP
                //  ---

                if (!DisabledMoves()) DoJump();
            }
            if (KB.ActionHeld || (isAiming && KB.MouseDown))
            {
                //  ---
                //  Keys ACTION
                //  ---

                if (!isFiring)
                {
                    // Check if held item is something aimable
                    if (!LB["LowerArmFront"].isActiveAndEnabled || !LB["LowerArmFront"].GripBehaviour.isHolding)
                    {
                        isFiring = false;
                        return;
                    }

                    ItemFH = LB["LowerArmFront"].GripBehaviour.CurrentlyHolding;

                    isFiring = true;

                    FireWeapon();
                }
                else if (useAuto) FireWeapon();
            }
        }


        public void doEmote()
        {
            if (!isEmote)
            {

                EmoteId = -1;

                KB.DisableNumberKeys();

                isEmote = true;
            }

            EmoteId = KB.CheckNumberKey();

            if (EmoteId == -1) return;

            string emotePose;

            switch (EmoteId)
            {
                case 1:
                    emotePose = "JT_Waving";
                    break;

                case 2:
                    emotePose = "JT_Victory";
                    break;

                case 3:
                    emotePose = "JT_Prone";
                    break;

                case 4:
                    emotePose = "JT_Dive_1";
                    break;

                case 5:
                    emotePose = "JT_Dive_2";
                    break;

                default:
                    emotePose = "JT_Dive_2";
                    break;
            }


            PBO.OverridePoseIndex = PuppetPose[emotePose];
        }
        private void FireWeapon()
        {
            if (useAuto)
            {
                ItemFH.SendMessage("UseContinuous", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
            }
            else
            {
                if (PBO.IsAlive()) { 
                    ItemFH.SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);
                } else isActive = false;
            }
        }


        public void StopPerson()
        {
            if (!isWalking && !isCrouching) { 

                if (PBO.OverridePoseIndex != (int)PoseState.Rest || PBO.OverridePoseIndex != (int)PoseState.Sitting)
                {
                    PBO.OverridePoseIndex = -1;
                }
            }
        }
        

        public void AimingInit()
        {
            // Check if held item is something aimable
            if (!LB["LowerArmFront"].isActiveAndEnabled || !LB["LowerArmFront"].GripBehaviour.isHolding)
            {
                isAiming = false;
                return;
            }

            //JTUtil.SnapShot(ItemFH.rigidbody, "mass,drag,inertia");
            // ItemFH.rigidbody.inertia = 0.001f;
            
            Type[] ShootersAuto = new Type[] { 
                typeof(FlamethrowerBehaviour),
                typeof(BeamformerBehaviour),
                typeof(LightningGunBehaviour),
                typeof(FireExtinguisherBehaviour),
            };

            Type[] ShootersManual = new Type[]
            {
                typeof(RocketLauncherBehaviour),
                typeof(GenericScifiWeapon40Behaviour),

            };

            Type[] ShootersCheck = new Type[]
            {
                typeof(FirearmBehaviour),
                typeof(CanShoot),
                typeof(ProjectileLauncherBehaviour),

            };
            
            ItemFH = LB["LowerArmFront"].GripBehaviour.CurrentlyHolding;
            
            bool canShoot = false;

            foreach (Type type in ShootersAuto)
            {
                if (ItemFH.TryGetComponent(type, out _))
                {
                    canShoot = true;
                    useAuto  = true;
                    return;
                }
            }

            foreach (Type type in ShootersManual)
            {
                if (ItemFH.TryGetComponent(type, out _))
                {
                    canShoot = true;
                    useAuto  = false;
                    return;
                }
            }

            //  Determine if we can do auto firing
            if (ItemFH.TryGetComponent<FirearmBehaviour>(out FirearmBehaviour FBH))
            {
                canShoot             = true;
                useAuto              = FBH.Automatic;
                FBH.Cartridge.Recoil = 0.01f;

            } else if (ItemFH.TryGetComponent<ProjectileLauncherBehaviour>(out ProjectileLauncherBehaviour PLB))
            {
                canShoot             = true;
                useAuto              = PLB.IsAutomatic;
                PLB.recoilMultiplier = 0.1f;

            } else
            {
                canShoot = false;
                useAuto  = false;
                canShoot = false;
            }

            isAiming = canShoot;
            
        }

        public void AimingStop()
        {
            LB["LowerArmFront"].Broken = false;
            LB["UpperArmFront"].Broken = false;
            //if (ItemFH != null) JTUtil.SnapShotRestore(ItemFH.rigidbody);
        }

        public bool altArm = true;

        public void AimWeapon()
        {
            //
            //  Reference: https://gamedevbeginner.com/make-an-object-follow-the-mouse-in-unity-in-2d/
            //
            Vector2 mouse = Camera.main.ScreenToWorldPoint(Input.mousePosition);

            if (facingLeft)
            {
                if (mouse.x > RB2["Head"].position.x) {
                    LB["LowerArmFront"].Broken = false;
                    LB["UpperArmFront"].Broken = false;
                    RB2["UpperArmFront"].inertia = 0;
                    RB2["LowerArmFront"].inertia = 0;

                    return;
                }
            }
            else
            {
                if (mouse.x < RB2["Head"].position.x) {
                    LB["LowerArmFront"].Broken = false;
                    LB["UpperArmFront"].Broken = false;
                    RB2["UpperArmFront"].inertia = 0;
                    RB2["LowerArmFront"].inertia = 0;
                    return;
                }
            }

            LB["UpperArmFront"].Broken = true;
            //LB["LowerArmFront"].Broken = true;
            //LB["LowerArmFront"].IsActiveInCurrentPose = false;
            //LB["UpperArmFront"].IsActiveInCurrentPose = false;
            //


            altArm = !altArm;

            RB2["UpperArmFront"].inertia = 0.00125f;
            RB2["LowerArmFront"].inertia = 0.00125f;

            Rigidbody2D rbarm = isWalking ? RB2["LowerArmFront"] : RB2["LowerArmFront"];
            Vector3 diff = ((Vector3)rbarm.position - Camera.main.ScreenToWorldPoint(Input.mousePosition));

            float facing = facingLeft ? 1 : -1;
            
            float rot_z = Mathf.Atan2(diff.y, diff.x) * Mathf.Rad2Deg;
            rbarm.MoveRotation(Quaternion.Euler(0f, 0f, rot_z - 90) );

        }


        public void DoJump(bool launch=false)
        {
            if (!launch) { 
                    
                if (!isJumping) { 
                    hasJumped                 = false;
                    PBO.OverridePoseIndex     = PuppetPose["JT_FrontFlip"];
                    isJumping                 = true;
                    JumpStrength              = 0;
                    TimeJump                  = Time.time;
                    
                    return;
                }

            } else
            {
                Vector2 direction       = new Vector2(KB.Right ? 3 : KB.Left ? -3 : 1 * (facingLeft ? -1f : 1f), 10);
                float torque = 0f;

                if (!hasJumped && PBO.IsTouchingFloor)
                {
                    hasJumped             = true;
                    PBO.OverridePoseIndex = -1;
                    JumpStrength          = (Time.time - TimeJump) * 2.1f;
                    TimeJump              = Time.time;


                    if (JumpStrength > 3) JumpStrength = 3;

                    if (KB.Left || KB.Right)
                    {
                        torque = (KB.Left ? 1 : -1) * 20 * TotalWeight * JumpStrength;
                        RB2["MiddleBody"].AddTorque(torque, ForceMode2D.Force);
                        RB2["UpperBody"].AddTorque(torque, ForceMode2D.Force);
                        RB2["LowerBody"].AddTorque(torque, ForceMode2D.Force);
                    }


                    RB2["UpperBody"].AddRelativeForce(direction * (10 * (JumpStrength * TotalWeight)));
                    RB2["LowerBody"].AddRelativeForce(direction * (10 * (JumpStrength * TotalWeight)));

                } else
                {
                    hasJumped        = true;

                    if (Time.time - TimeJump < 0.1f) {

                        RB2["LowerBody"].AddForce(direction * (JumpStrength * TotalWeight));

                    } else if (PBO.IsTouchingFloor)
                    {
                        PBO.OverridePoseIndex   = -1;
                        isJumping               = false;
                        DisableMoves            = Time.time + JumpStrength / 2;
                        RestoreLimbSnapshot();
                        return;
                    }

                    if (KB.Left)
                    {
                        RB2["Head"].AddForce((RB2["Head"].transform.right * -1) * TotalWeight * JumpStrength);
                        torque = 1f * TotalWeight * JumpStrength * 0.1f;
                        RB2["MiddleBody"].AddTorque(torque, ForceMode2D.Force);
                        RB2["UpperBody"].AddTorque(torque, ForceMode2D.Force);
                        RB2["LowerBody"].AddTorque(torque, ForceMode2D.Force);
                    }
                    else if (KB.Right)
                    {
                        RB2["Head"].AddForce(RB2["Head"].transform.right * TotalWeight * JumpStrength);
                        torque = -1f * TotalWeight * JumpStrength * 0.1f;
                        RB2["MiddleBody"].AddTorque(torque, ForceMode2D.Force);
                        RB2["UpperBody"].AddTorque(torque, ForceMode2D.Force);
                        RB2["LowerBody"].AddTorque(torque, ForceMode2D.Force);
                    }

                }
            }
        }

        public void DoCrouch()
        {
            if (KB.KeyCombos.DoubleDown)
            {
                Actions.prone.Init();
            }
            else
            {
                PBO.OverridePoseIndex  = PuppetPose["JT_Duck"];


            }
            isCrouching = true; 
        }

        

        public void DoThrow(bool launch=false)
        {

            if (isAiming) return;
            if (!isThrowing && doneThrowing)
            {
                Holding = Thing.HeldThing(PBO);

                if (Holding == null)
                {
                    isThrowing = false;
                    return;
                }
                doneThrowing = false;
                isThrowing   = true;
                ThrowWeight  = Holding.P.InitialMass;
                ThrowForce   = 0;

                PBO.OverridePoseIndex = PuppetPose["JT_Throw_x"];

                TimeAnimation = Time.time;

                return;

            } 
            else if (launch)
            {
                ThrowForce = Time.time - TimeAnimation;

                if (ThrowForce > 3) ThrowForce = 3;

                PBO.OverridePoseIndex = -1;

                //ThrowHeldItem(ThrowForce);

                isThrowing = false;

                TimeAnimation = Time.time + 0.15f;

                RB2["UpperArmFront"].AddTorque(15f * (1f + RB2["UpperArmFront"].mass) * (facingLeft ? 1 : -1));
                RB2["LowerArmFront"].AddTorque(25f * (1f + RB2["LowerArmFront"].mass) * (facingLeft ? 1 : -1));

            }

        }

        public void ThrowHeldItem(float force)
        {

            PuppetMaster.ChaseCam.QuickChase(Holding, 5);

            RB2["UpperArmFront"].AddTorque(25f * (1f + RB2["UpperArmFront"].mass) * (facingLeft ? 1 : -1));
            Holding.BreakConnections();

            Vector3 diff = (Camera.main.ScreenToWorldPoint(Input.mousePosition) - (Vector3)Holding.R.position);
            
            Holding.P.MakeWeightful();
            Holding.R.AddForce(diff * (ThrowForce * 50) * ThrowWeight);
            Holding.R.AddForceAtPosition(Vector2.up * (ThrowForce * 50) * ThrowWeight, RB2["LowerArmFront"].position);

            JTUtil.Notify("Threw Item");

            Collider2D[] noCollide = GripF.transform.root.GetComponentsInChildren<Collider2D>();

            foreach (Collider2D col1 in Holding.G.transform.root.GetComponentsInChildren<Collider2D>())
            {
                foreach (Collider2D col2 in noCollide)
                {
                    if ((bool)(UnityEngine.Object)col2 && (bool)(UnityEngine.Object)col1)
                        Physics2D.IgnoreCollision(col1, col2);
                }
            }
            
        }
        
        public void DoPickup(PhysicalBehaviour PBItem=null)
        {
            Collider2D[] noCollide;
            PhysicalBehaviour phys;
            
            if (PBItem != null)     phys = PBItem;
            else                    phys = JTUtil.CanHold( PuppetMaster.LastClickedItem ) 
                                         ? PuppetMaster.LastClickedItem 
                                         : null;

            Rigidbody2D LowerArm    = RB2["LowerArmFront"];
            Rigidbody2D AltArm      = LowerArm == RB2["LowerArm"] ? RB2["LowerArmFront"] : RB2["LowerArm"];

            if( !LowerArm ) return;

            float ArmRotation   = LowerArm.rotation;

            GripBehaviour GB    = null;    
            
            LowerArm.TryGetComponent<GripBehaviour>(out GB);

            if (GB.isHolding)
            {
                //  DROP ITEM 
                PhysicalBehaviour physo = GB.CurrentlyHolding;

                noCollide = GB.transform.root.GetComponentsInChildren<Collider2D>();

                LowerArm.SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);

                foreach (Collider2D col1 in physo.transform.root.GetComponentsInChildren<Collider2D>())
                {
                    foreach (Collider2D col2 in noCollide)
                    {
                        if ((bool)(UnityEngine.Object)col2 && (bool)(UnityEngine.Object)col1)
                            Physics2D.IgnoreCollision(col1, col2);
                    }
                }

                //  Add objects to list for re-enabling colliders
                //  DisabledCollisions.Add(new DisabledCollision(PBO.transform, physo));

                // CheckDisabledCollisions     = true;
                FixedJoint2D ItemjointStuck = null;
                GB.isHolding                = false;

                physo.MakeWeightful();

                while (GB.gameObject.TryGetComponent<FixedJoint2D>(out FixedJoint2D ItemJoint))
                {
                    if (ItemjointStuck == ItemJoint) break; 
                    DestroyImmediate(ItemJoint);
                    ItemjointStuck = ItemJoint;
                }

                //if (CustomPoses.ContainsKey(PBO.ActivePose.Name)) PBO.OverridePoseIndex = -1;
                isHolding   = false;
                ItemFH      = null;

                return;
            }

            Vector2 worldPoint        = (Vector2)GB.transform.TransformPoint(GB.GripPosition);
            Vector2 NearestHoldingPos = new Vector2(0.0f, 0.0f);

            if (!(bool)(UnityEngine.Object)phys) return;

            bool ItemFlipped = phys.transform.localScale.x < 0.0f;

            if (isFlipped() != ItemFlipped) JTUtil.FlipItem(phys);
            
            ItemFH = phys;


            phys.transform.rotation = Quaternion.Euler(0.0f, 0.0f, isFlipped() ? ArmRotation + 95.0f : ArmRotation - 95.0f);

            //  Adjust layers of the hand and the held object so object is over body but under hand and arms
            int SOrder = LowerArm.GetComponent<SpriteRenderer>().sortingOrder;
            if (SOrder <= 1)
            {
                SOrder = 2;
                LowerArm.GetComponent<SpriteRenderer>().sortingOrder = SOrder;
                PBO.transform.Find("FrontArm").transform.Find("UpperArmFront").GetComponent<SpriteRenderer>().sortingOrder = SOrder;
            }

            phys.GetComponent<SpriteRenderer>().sortingLayerName = LowerArm.GetComponent<SpriteRenderer>().sortingLayerName;
            phys.GetComponent<SpriteRenderer>().sortingOrder     = SOrder + -1;

            //  The majority of items seem to look better using the grippos with the lower X val when not flipped
            foreach (Vector3 HPos in phys.HoldingPositions) if (HPos.x < NearestHoldingPos.x) NearestHoldingPos = HPos;

            phys.transform.position += GB.transform.TransformPoint(GB.GripPosition) - phys.transform.TransformPoint((Vector3)NearestHoldingPos);
            phys.MakeWeightless();

            LowerArm.SendMessage("Use", (object)new ActivationPropagation(), SendMessageOptions.DontRequireReceiver);

            //  Disables collisions between the object and person holding it (also alt hand item).
            noCollide = GB.transform.root.GetComponentsInChildren<Collider2D>();

            if (AltArm.TryGetComponent<GripBehaviour>(out GripBehaviour GBAlt))
            {
                if (GBAlt.isHolding)
                {
                    PhysicalBehaviour PBAltItem = GBAlt.CurrentlyHolding;

                    foreach (Collider2D col1 in PBAltItem.transform.root.GetComponentsInChildren<Collider2D>())
                    {
                        foreach (Collider2D col2 in phys.transform.root.GetComponentsInChildren<Collider2D>())
                        {
                            if ((bool)(UnityEngine.Object)col2 && (bool)(UnityEngine.Object)col1)
                                Physics2D.IgnoreCollision(col1, col2);
                        }
                    }
                }
            }

            isHolding = true;
        }


        public void SetLimbSnapshot()
        {
            LimbSnapshots.Clear();

            foreach (KeyValuePair<string, LimbBehaviour> kvp in LB)
            {
                LimbSnapshot snapshot = new LimbSnapshot()
                {
                    Health            = kvp.Value.Health,
                    BreakingThreshold = kvp.Value.BreakingThreshold,
                    Broken            = kvp.Value.Broken,
                    BruiseCount       = kvp.Value.BruiseCount,
                    Frozen            = kvp.Value.Frozen,
                    Numbness          = kvp.Value.Numbness,
                    BaseStrength      = kvp.Value.BaseStrength,
                    RegenerationSpeed = kvp.Value.RegenerationSpeed,
                    Vitality          = kvp.Value.Vitality,
                };

                LimbSnapshots.Add(kvp.Key, snapshot);
            }
        }

        public void Strengthen()
        {
            foreach (KeyValuePair<string, LimbBehaviour> kvp in LB)
            {
                LB[kvp.Key].Vitality             *= 0.1f;
                LB[kvp.Key].RegenerationSpeed    += 5f;
                LB[kvp.Key].ImpactPainMultiplier *= 0.1f;
                LB[kvp.Key].InitialHealth        += 150f;
                LB[kvp.Key].BreakingThreshold    += 10f;
                LB[kvp.Key].ShotDamageMultiplier *= 0.1f;
                LB[kvp.Key].Health               += LB[kvp.Key].InitialHealth;
                LB[kvp.Key].BaseStrength          = Mathf.Min(15f, LB[kvp.Key].BaseStrength + 5f);
            }
        }

        public void RestoreLimbSnapshot(bool mend=false)
        {
            foreach (KeyValuePair<string, LimbSnapshot> kvp in LimbSnapshots)
            {
                LB[kvp.Key].Health            = kvp.Value.Health;
                LB[kvp.Key].BreakingThreshold = kvp.Value.BreakingThreshold;
                LB[kvp.Key].Broken            = kvp.Value.Broken;
                LB[kvp.Key].BruiseCount       = kvp.Value.BruiseCount;
                LB[kvp.Key].Frozen            = kvp.Value.Frozen;
                LB[kvp.Key].Numbness          = kvp.Value.Numbness;
                LB[kvp.Key].BaseStrength      = kvp.Value.BaseStrength;
                LB[kvp.Key].RegenerationSpeed = kvp.Value.RegenerationSpeed;
                LB[kvp.Key].Vitality          = kvp.Value.Vitality;


                if (mend)
                {
                    LB[kvp.Key].Health = LB[kvp.Key].InitialHealth;
                    LB[kvp.Key].CirculationBehaviour.BleedingRate  *= 0.8f;
                    LB[kvp.Key].SkinMaterialHandler.RottenProgress *= 0.8f;
                    LB[kvp.Key].HealBone();
                }
                
            }
        }

        

        public void OnDestroy()
        {
            PuppetMaster.DeactivatePuppet(this.GetHashCode());
        }
    }
}
