﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.Internal;
using UnityEngine.EventSystems;

namespace JTShtuff
{  
    public class PuppetMaster : MonoBehaviour
    {
        public static bool Enabled                                    = false;
        public static int ActivePuppet                  { get; set; } = 0;
        public static PhysicalBehaviour LastClickedItem { get; set; } = null;
        public static PersonBehaviour LastClickedPerson { get; set; } = null;
        public static PuppetBehaviour Puppet            { get; set; } = null;
        public static ChaseCam ChaseCam                 { get; set; } = null;
        public static CustomPoses JTPoses               { get; set; } = new CustomPoses();

        public static bool CanBePuppet(PersonBehaviour person) => (bool) (person.isActiveAndEnabled && person.IsAlive());


        public void Start()
        {
            KB.SwapBindings(Enabled);
        }

        public void Update()
        {
            if (InputSystem.Down("PM-TogglePM")) TogglePuppetMode();

            if (!Enabled) return;

            KB.CheckKeys();

            if (KB.PClick) CheckNewClicks();

        }


        public static void TogglePuppetMode()
        {
            Enabled = !Enabled;

            KB.SwapBindings(Enabled);

            if (Enabled && ChaseCam == null)
            {
                Global.main.gameObject.GetOrAddComponent<ChaseCam>();

                // Create chase-cam object
                //Texture2D tex            = new Texture2D(0, 0);
                //Sprite mySprite          = Sprite.Create(tex, new Rect(0.0f, 0.0f, tex.width, tex.height), new Vector2(0.1f, 0.1f), 0.1f);
                //GameObject ChaseCamObj   = ModAPI.CreatePhysicalObject("ChaseCam", mySprite);
                //ChaseCam                 = ChaseCamObj.GetOrAddComponent<ChaseCam>();
            }

            if (Enabled && CustomPoses.Poses.Count == 0)
            {
                // Init the custom poses
                CustomPoses.InitCustomPoses();
            }

            JTUtil.Notify("Puppet Master: <color=" + (Enabled ? "yellow>On" : "red>Off") + "</color>", JTUtil.VerboseLevels.Minimal);

        }


        public static void CheckNewClicks()
        {
            PersonBehaviour LCP = JTUtil.GetClickedPerson();
            if (LCP != null)
            {

                if (LCP == LastClickedPerson)
                {
                    //  This person was clicked twice, so make them the puppet
                    if (CanBePuppet(LCP)) ActivatePuppet(LCP);
                    
                    LCP = null;

                    return;

                }

                LastClickedPerson = LCP;
            }
            else
            {
                PhysicalBehaviour LCI = JTUtil.GetClickedItem();
                if (LCI != null)
                {

                    if (LCI == LastClickedItem)
                    {

                        //  Item was clicked twice, so pick it up automatically
                        if (Puppet != null && Puppet.isActive ) Puppet.DoPickup();

                        return;
                    }

                    LastClickedItem = LCI;

                }
            }
        }

        public static void ActivatePuppet(PersonBehaviour puppet, bool deactivateLast = true)
        {
            int puppetId = puppet.GetHashCode();

            if (ActivePuppet != puppetId) DeactivatePuppet(ActivePuppet);
            

            PuppetBehaviour MyPuppet =  puppet.gameObject.GetOrAddComponent<PuppetBehaviour>();
            
            if (MyPuppet != null) {

                if (ActivePuppet == puppetId) {
                    
                    //  This is already the puppet, so refresh some settings
                    ChaseCam.Chase(
                        Puppet.GetComponentInChildren<PhysicalBehaviour>(),
                        MyPuppet.GetHashCode()
                    );
                }

                MyPuppet.isActive      = true;
                Puppet                 = MyPuppet;
                ActivePuppet           = puppetId;
            }
        }


        public static void DeactivatePuppet(int puppetHash)
        {
            PersonBehaviour person = JTUtil.FindPeepsByHash(puppetHash); 
            
            ChaseCam.StopChase(puppetHash);

            if (person != null) DeactivatePuppet(person);

        }

        public static void DeactivatePuppet(PersonBehaviour puppet)
        {
            if (puppet.TryGetComponent<PuppetBehaviour>(out PuppetBehaviour jtc))
            {
                UnityEngine.Object.Destroy((UnityEngine.Object)jtc);
            }

            if (puppet.TryGetComponent<InventoryBehaviour>(out InventoryBehaviour IB))
            {
                IB.Enabled = false;
            }

        }

        
    }
}
