﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using Unity;
using UnityEngine;

namespace JTShtuff
{
    public class Thing
    {
        public GameObject G;
        public PhysicalBehaviour P;
        public Rigidbody2D R;
        public string Name;
        public int Hash;

        public Thing(GameObject gameObject)
        {
            G    = gameObject;
            P    = gameObject.GetComponent<PhysicalBehaviour>();
            R    = gameObject.GetComponent<Rigidbody2D>();
            Name = P.name;
            Hash = P.GetHashCode();
        }

        public Thing(PhysicalBehaviour physicalBehaviour)
        {
            P    = physicalBehaviour;
            G    = physicalBehaviour.gameObject;
            R    = G.GetComponent<Rigidbody2D>();
            Name = P.name;
            Hash = P.GetHashCode();
        }

        public static Thing HeldThing(PersonBehaviour PBO)
        {
            GripBehaviour[] Grips = PBO.GetComponentsInChildren<GripBehaviour>();

            foreach (GripBehaviour grip in Grips)
            {
                if(grip.isHolding) return new Thing(grip.CurrentlyHolding);
            }

            return (Thing) null;
        }

        public PersonBehaviour HeldBy()
        {
            if (!P.beingHeldByGripper) return (PersonBehaviour) null;

            GripBehaviour[] Grips = Global.FindObjectsOfType<GripBehaviour>();

            foreach (GripBehaviour grip in Grips)
            {
                if (grip != null && grip.CurrentlyHolding == P)
                {
                    return grip.GetComponentInParent<PersonBehaviour>();
                }
            }

            return (PersonBehaviour) null;
        }

        public void BreakConnections()
        {
            if (P.beingHeldByGripper)
            {
                PersonBehaviour PBO = HeldBy();

                if (PBO != null)
                {
                    FixedJoint2D ItemjointStuck = null;

                    GripBehaviour[] Grips = PBO.GetComponentsInChildren<GripBehaviour>();

                    foreach (GripBehaviour grip in Grips)
                    {
                        if (grip.CurrentlyHolding != P) continue;

                        grip.DropObject();
                        
                    }
                }
            }
        }

        public void Activate()
        {

        }

        public void Destroy()
        {
            BreakConnections();

            if (P.Deletable)
            {
                G.SendMessage("OnUserDelete", SendMessageOptions.DontRequireReceiver);

                UnityEngine.Object.Destroy((UnityEngine.Object)G);
            }

            P = (PhysicalBehaviour)null;
            R = (Rigidbody2D)null;
            G = (GameObject)null;
        }
    }
}
