﻿//     ___                        _                    _     
//    / _ \_   _ _ __  _ __   ___| |_  /\/\   __ _ ___| |_ ___ _ __ 
//   / /_)/ | | | '_ \| '_ \ / _ \ __|/    \ / _` / __| __/ _ \ '__|
//  / ___/| |_| | |_) | |_) |  __/ |_/ /\/\ \ (_| \__ \ ||  __/ |  
//  \/     \__,_| .__/| .__/ \___|\__\/    \/\__,_|___/\__\___|_|  
//              |_|   |_|                    My code is your code  
//                                           
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Unity;
using UnityEngine;
using UnityEngine.EventSystems;
using UnityEngine.Events;

namespace JTShtuff
{
    public static class JTUtil
    {
        private static Dictionary<int, List<SnapshotItems>> Snapshots = new Dictionary<int, List<SnapshotItems>>();
        public static void Destroy(UnityEngine.Object o) => UnityEngine.Object.Destroy((UnityEngine.Object) o);
        public enum VerboseLevels
        {
            Off,
            Minimal,
            Full,
        };

        public struct SnapshotItems
        {
            public string propName;
            public object propValue;
        }

        

        public static void Notify( string msg, VerboseLevels verboseLevel=VerboseLevels.Minimal)
        {
            if ((int)JTMain.Verbose >= (int)verboseLevel) ModAPI.Notify( msg );
        }

        public static PersonBehaviour FindPeepsByHash(int hashId)
        {
            PersonBehaviour[] people = Global.FindObjectsOfType<PersonBehaviour>();

            for (int i = 0; i < people.Length; i++)
            {
                if (people[i].GetHashCode() == hashId) return people[i];
            }

            return null;
        }


        public static PersonBehaviour GetClickedPerson()
        {
            foreach (PhysicalBehaviour PB in Global.main.PhysicalObjectsInWorld)
            {
                if (!PB.Selectable || PB.colliders.Length == 0) continue;

                foreach (Collider2D collider in PB.colliders)
                {
                    if (!collider) continue;

                    if (collider.OverlapPoint((Vector2)Global.main.MousePosition))
                    {
                        return PB.GetComponentInParent<PersonBehaviour>();
                    }
                }
            }

            return null;
        }


        public static PhysicalBehaviour GetClickedItem()
        {
            foreach (PhysicalBehaviour PB in Global.main.PhysicalObjectsInWorld)
            {
                if (!PB.Selectable || PB.colliders.Length == 0) continue;

                foreach (Collider2D collider in PB.colliders)
                {
                    if (collider && collider.OverlapPoint((Vector2)Global.main.MousePosition)) return PB;
                }
            }

            return null;
        }


        public static void FlipItem(PhysicalBehaviour PB)
        {
            Vector3 theScale = PB.transform.localScale;
            
            theScale.x *= -1;
            
            PB.transform.localScale = theScale;
        }

        //public struct hpos
        //{
        //    public Vector3 GripPoint;
        //    public Vector3 

        //}

        public static bool FlipPerson(PersonBehaviour PBO, bool okSpasticated=false)
        {

            if (!okSpasticated && (!PBO.isActiveAndEnabled || !PBO.IsAlive() || !PBO.IsTouchingFloor || PBO.Consciousness < 1)) return false;

            Flipper flipper;
            Rigidbody2D head;
            Vector2 moveDif;
            PhysicalBehaviour hObj = null;
            FixedJoint2D Itemjoint;
            FixedJoint2D joint;
            Vector2 GripPoint;



            List<Vector3> PreviousGrips = new List<Vector3>();

            GripBehaviour[] preGrips = PBO.GetComponentsInChildren<GripBehaviour>();

            foreach (GripBehaviour grip in preGrips)
            {
                if (grip.isHolding)
                {
                    GripPoint = grip.CurrentlyHolding.GetNearestLocalHoldingPoint(grip.transform.TransformPoint(grip.GripPosition), out float distance);
                    PreviousGrips.Add(GripPoint);
                    //Vector3 hp = grip.transform.TransformPoint(grip.GripPosition) - grip.CurrentlyHolding.transform.TransformPoint(GripPoint);
                    //hObj = grip.CurrentlyHolding;
                }
            }

            
            
            flipper = new Flipper()
            {
                PB = PBO,
                flipScale = PBO.transform.localScale,
                instanceID = PBO.GetInstanceID(),
                status = 1
            };

            flipper.flipScale.x *= -1;

            foreach (LimbBehaviour limb in PBO.Limbs)
            {
                if (limb == PBO.Limbs[1]) flipper.moveB = limb.transform.position;
                if (limb.HasJoint)
                {
                    limb.BreakingThreshold *= 8;

                    if (limb.name != "LowerBody"
                        && limb.name != "MiddleBody"
                        && limb.name != "UpperBody"
                        && limb.name != "UpperArm"
                        && limb.name != "UpperArmFront"
                        && limb.name != "Head")
                    {
                        JointAngleLimits2D t = limb.Joint.limits;
                        t.min *= -1f;
                        t.max *= -1f;
                        limb.Joint.limits = t;
                        limb.OriginalJointLimits = new Vector2(limb.OriginalJointLimits.x * -1f, limb.OriginalJointLimits.y * -1f);
                    }
                }
            }

            Transform headT = PBO.transform.Find("Head");

            if (headT)
            {
                head = headT.GetComponent<Rigidbody2D>();

                PBO.transform.localScale = flipper.flipScale;



                foreach (LimbBehaviour limb in PBO.Limbs)
                {
                    if (limb == PBO.Limbs[1]) flipper.moveA = head.transform.position;
                }




                moveDif = flipper.moveB - flipper.moveA;
                PBO.AngleOffset *= -1f;
                PBO.transform.position = new Vector2(PBO.transform.position.x + moveDif.x, PBO.transform.position.y);

                foreach (LimbBehaviour limb in PBO.Limbs) if (limb.HasJoint) limb.Broken = false;
                GripBehaviour[] grips = PBO.GetComponentsInChildren<GripBehaviour>();
                bool alreadyFlipped = false;
                int gcount = 0;
                foreach (GripBehaviour grip in grips)
                {
                    if (grip.isHolding)
                    {
                        alreadyFlipped = (hObj && hObj == grip.CurrentlyHolding);
                        hObj = grip.CurrentlyHolding;

                        // Break joint
                        while (Itemjoint = grip.gameObject.GetComponent<FixedJoint2D>())
                        {
                            UnityEngine.Object.DestroyImmediate(Itemjoint);
                        }

                        
                        
                        if (!alreadyFlipped)
                        {
                            //  Flip Item
                            Vector3 theScale = hObj.transform.localScale;
                            theScale.x *= -1.0f;
                            hObj.transform.localScale = theScale;

                            //  Set new item rotation
                            hObj.transform.rotation = Quaternion.Euler(
                                0.0f, 0.0f,
                                grip.GetComponentInParent<Rigidbody2D>().rotation + 95.0f * (PBO.transform.localScale.x < 0.0f ? 1.0f : -1.0f)
                            );

                            //  Move item to flipped position
                            //GripPoint = hObj.GetNearestLocalHoldingPoint(grip.transform.TransformPoint(grip.GripPosition), out float distance);

                            GripPoint = PreviousGrips[gcount];
                            gcount++;
                            hObj.transform.position += grip.transform.TransformPoint(grip.GripPosition) -
                                hObj.transform.TransformPoint(GripPoint);

                            //  Create new joint
                            joint                 = grip.gameObject.AddComponent<FixedJoint2D>();
                            joint.connectedBody   = hObj.rigidbody;
                            joint.anchor          = (Vector2)grip.GripPosition;
                            joint.connectedAnchor = GripPoint;
                            joint.enableCollision = false;

                        }
                        else
                        {
                            GripPoint = hObj.GetNearestLocalHoldingPoint(grip.transform.TransformPoint(grip.GripPosition), out float distance);
                            hObj.transform.position += grip.transform.TransformPoint(grip.GripPosition) -
                                hObj.transform.TransformPoint(GripPoint);
                            //  Create new joint
                            joint                 = grip.gameObject.AddComponent<FixedJoint2D>();
                            joint.connectedBody   = hObj.rigidbody;
                            joint.anchor          = (Vector2)grip.GripPosition;
                            joint.connectedAnchor = GripPoint;
                            joint.enableCollision = false;

                        }


                    }
                }
            }
            return true;
        }

        //public static bool FlipPersonx(PersonBehaviour PBO, bool okSpasticated=false)
        //{
        //    if (!okSpasticated && (!PBO.isActiveAndEnabled || !PBO.IsAlive() || !PBO.IsTouchingFloor || PBO.Consciousness < 1)) return false;


        //    PhysicalBehaviour hObj = null;

            
        //    public Vector2 moveA;
        //    public Vector2 moveB;
        //    public Vector3 flipScale;

        //    GripBehaviour[] gripsx = PBO.GetComponentsInChildren<GripBehaviour>();

        //    PBO.transform.localscale.flipScale.x *= -1;

        //    foreach (LimbBehaviour limb in PBO.Limbs)
        //    {
        //        if (limb == PBO.Limbs[1]) moveB = limb.transform.position;
        //        if (limb.HasJoint)
        //        {
        //            limb.BreakingThreshold *= 8;

        //            if (limb.name != "LowerBody"
        //                && limb.name != "MiddleBody"
        //                && limb.name != "UpperBody"
        //                && limb.name != "UpperArm"
        //                && limb.name != "UpperArmFront"
        //                && limb.name != "Head")
        //            {
        //                JointAngleLimits2D t = limb.Joint.limits;
        //                t.min *= -1f;
        //                t.max *= -1f;
        //                limb.Joint.limits = t;
        //                limb.OriginalJointLimits = new Vector2(limb.OriginalJointLimits.x * -1f, limb.OriginalJointLimits.y * -1f);
        //            }
        //        }
        //    }

        //    Transform headT = PBO.transform.Find("Head");

        //    if (!headT || !headT.TryGetComponent<Rigidbody2D>(out Rigidbody2D head)) return false;

        //    PBO.transform.localScale = flipScale;

        //    foreach (LimbBehaviour limb in PBO.Limbs) 
        //        if (limb == PBO.Limbs[1]) moveA = head.transform.position;
            
        //    Vector2 moveDif = moveB - moveA;
            
        //    PBO.AngleOffset *= -1f;
        //    PBO.transform.position = new Vector2(PBO.transform.position.x + moveDif.x, PBO.transform.position.y);

        //    foreach (LimbBehaviour limb in PBO.Limbs) if (limb.HasJoint) limb.Broken = false;

        //    GripBehaviour[] grips = PBO.GetComponentsInChildren<GripBehaviour>();
        //    bool alreadyFlipped   = false;

        //    foreach (GripBehaviour grip in grips)
        //    {
        //        if (grip.isHolding)
        //        {
        //            FixedJoint2D joint;
        //            Vector2 GripPoint;

        //            alreadyFlipped = (hObj && hObj == grip.CurrentlyHolding);
        //            hObj           = grip.CurrentlyHolding;

        //            FixedJoint2D Itemjoint;

        //            // Break joint
        //            while (Itemjoint = grip.gameObject.GetComponent<FixedJoint2D>())
        //            {
        //                UnityEngine.Object.DestroyImmediate(Itemjoint);
        //            }


        //            if (!alreadyFlipped)
        //            {
        //                //  Flip Item
        //                Vector3 theScale          = hObj.transform.localScale;
        //                theScale.x               *= -1.0f;
        //                hObj.transform.localScale = theScale;

        //                //  Set new item rotation
        //                hObj.transform.rotation = Quaternion.Euler( 0.0f, 0.0f,
        //                    grip.GetComponentInParent<Rigidbody2D>().rotation + 95.0f * (PBO.transform.localScale.x < 0.0f ? 1.0f : -1.0f)
        //                );

        //                //
        //                // ─── TOO BUGGY AFTER FLIP ───────────────────────────────────────────────────────
        //                //
        //                //  Move item to flipped position
        //                //  @TODO: Learn the math to do this properly and scrap the tardcode
        //                //
        //                GripPoint = hObj.GetNearestLocalHoldingPoint(grip.transform.TransformPoint(grip.GripPosition), out float distance);

        //                hObj.transform.position  += grip.transform.TransformPoint(grip.GripPosition) - hObj.transform.TransformPoint(GripPoint);
        //                FixedJoint2D[] heldJoints = grip.gameObject.GetComponents<FixedJoint2D>();

        //                joint                 = grip.gameObject.AddComponent<FixedJoint2D>();
        //                joint.connectedBody   = hObj.rigidbody;
        //                joint.anchor          = (Vector2)grip.GripPosition;
        //                joint.connectedAnchor = GripPoint;
        //                joint.enableCollision = false;

        //                //for (int i = heldJoints.Length; --i >= 0;)
        //                //{
        //                //    heldJoints[i].connectedBody        = hObj.rigidbody;
        //                //    heldJoints[i].anchor               = (Vector2)grip.GripPosition;
        //                //    heldJoints[i].connectedAnchor      = GripPoint;
        //                //    heldJoints[i].transform.position   = joint.transform.position;
        //                //    heldJoints[i].transform.rotation   = joint.transform.rotation;
        //                //    heldJoints[i].transform.localScale = joint.transform.localScale;
        //                //}

        //                //UnityEngine.Object.DestroyImmediate(joint);


        //                //
        //                // ─── THIS VERSION CAUSES ISSUES WITH DROPPING AN ITEM USING THE "USE" ACTION ───────────────────────────────────────────────────────
        //                //
        //                //  Move item to flipped position


        //            }
        //            else
        //            {
        //                GripPoint = hObj.GetNearestLocalHoldingPoint(grip.transform.TransformPoint(grip.GripPosition), out float distance);
        //                hObj.transform.position += grip.transform.TransformPoint(grip.GripPosition) -
        //                hObj.transform.TransformPoint(GripPoint);
                        
        //                //  Create new joint
        //                joint                 = grip.gameObject.AddComponent<FixedJoint2D>();
        //                joint.connectedBody   = hObj.rigidbody;
        //                joint.anchor          = (Vector2)grip.GripPosition;
        //                joint.connectedAnchor = GripPoint;
        //                joint.enableCollision = false;

        //            }
        //        }
        //    }

        //    return true;
        //}

        public static bool CanHold(PhysicalBehaviour PB)
        {
            if (PB == null) return false;

            if (PB.beingHeldByGripper || PB.TryGetComponent<FreezeBehaviour>(out _)) return false;

            if (PB.HoldingPositions.Length < 1) return false;

            bool canHoldItem = false;

            SpawnableAsset SpawnName = CatalogBehaviour.Main.GetSpawnable(PB.name);

            if (SpawnName)
            {
                switch (SpawnName.Category.ToString().ToLower())
                {
                    case "melee (category)":
                    case "firearms (category)":
                    case "explosives (category)":
                    case "biohazard (category)":
                        canHoldItem = true;
                        break;

                    default:
                        break;
                }
            }

            canHoldItem = true;

            return canHoldItem;

        }

        public static bool SnapShot(object obj, string delimList)
        {
            if (Snapshots.ContainsKey(obj.GetHashCode())) return false;

            string[] propsToSave = delimList.Split(',');

            List<SnapshotItems> list = new List<SnapshotItems>();

            foreach (string propName in propsToSave)
            {
                object propVal = obj.GetType().GetProperty(propName)?.GetValue(obj, null);

                if (propVal == null) continue;

                list.Add(new SnapshotItems() { propName = propName, propValue = propVal });
            }

            Snapshots.Add(obj.GetHashCode(), list);

            return true;
        }


        public static bool SnapShotRestore(object obj)
        {
            if (Snapshots.TryGetValue(obj.GetHashCode(), out List<SnapshotItems> PropList))
            {
                Snapshots.Remove(obj.GetHashCode());

                foreach (SnapshotItems item in PropList)
                {
                    obj.GetType().GetProperty(item.propName)?.SetValue(obj, item.propValue, null);
                }

                return true;
            }

            return false;
        }
        private struct Flipper
        {
            public int instanceID;
            public PersonBehaviour PB;
            public Vector2 moveA;
            public Vector2 moveB;
            public Vector3 flipScale;
            public int status;

            public override int GetHashCode() => base.GetHashCode();
            public override bool Equals(object obj) => (int)obj == instanceID;
        }

    }


}
